/*
 *  Copyright (c) xlightweb.org, 2008 - 2009. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;

import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xsocket.connection.IConnection.FlushMode;



/**
 * I/O resource capable of receiving the body data.
 * 
 * @author grro@xlightweb.org
 *
 */
abstract class AbstractNetworkBodyDataSink extends BodyDataSinkImplBase {

    private static final Logger LOG = Logger.getLogger(AbstractNetworkBodyDataSink.class.getName());

    private final AbstractHttpConnection httpConnection;
    
         
    /**
     * constructor 
     * 
     * @param httpConnection   the http connection
     */
    public AbstractNetworkBodyDataSink(AbstractHttpConnection httpConnection, String characterEncoding) throws IOException {
        super(characterEncoding, httpConnection.getExecutor());
        this.httpConnection = httpConnection;
        
        this.httpConnection.setNetworkBodyDataSink(this);
        httpConnection.getUnderlyingTcpConnection().setFlushmode(FlushMode.ASYNC);
        httpConnection.getUnderlyingTcpConnection().setAutoflush(false);
    }
    
    
    void onDisconnect() {
        httpConnection.setNetworkBodyDataSink(null);
        
        // data sink should not be open
        if (isOpen()) {
            destroy();
        }
    }
    

    
    
    /**
     * {@inheritDoc}
     */
    @Override
    protected boolean isNetworkendpoint() {
        return true;
    }
    
    /**
     * returns the http connection
     * @return the http connection 
     */
    final AbstractHttpConnection getConnection() {
        return httpConnection;
    }
 
    
    public final String getId() {
        return httpConnection.getId();
    }
    
    @Override
    final void onClose() throws IOException {
        
        httpConnection.removeNetworkBodyDataSink(this);
            
        try {
            performClose();
            httpConnection.onPostMessageWrite();
        } catch(IOException ioe) {
            
            if (isIgnoreWriteError()) {
                if (LOG.isLoggable(Level.FINE)) {
                    LOG.fine("[" + getId() + "] error occured by closing connection. ignoring it (isIgnoreWriteError=true) " + ioe.toString());
                }
                
            } else {
                if (LOG.isLoggable(Level.FINE)) {
                    LOG.fine("[" + getId() + "] error occured by closing connection. destroying it " + ioe.toString());
                }               
                destroy();
                throw ioe;
            }
        }
    }
    
    
    abstract void performClose() throws IOException;
    

    /**
     * {@inheritDoc}
     */
    @Override
    void onDestroy(String reason) {
        httpConnection.removeNetworkBodyDataSink(this);
        
        if (isOpen()) {
            if (LOG.isLoggable(Level.FINE)) {
                LOG.fine("[" + getId() + "] destroying connection");
            }
            
            try {
                performDestroy();
            } catch (IOException ioe) {
                if (LOG.isLoggable(Level.FINE)) {
                    LOG.fine(ioe.toString());
                }
            }
        }
        
        httpConnection.destroy(reason);
    }
    
    
    abstract void performDestroy() throws IOException;
    
    
    /**
     * {@inheritDoc}
     */
    public int getPendingWriteDataSize() {
        return httpConnection.getUnderlyingTcpConnection().getPendingWriteDataSize();
    }
    
    
    
    /**
     * 
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        if (isOpen()) {
            return getClass().getName() + "#" + hashCode();
        } else {
            return getClass().getName() + "#" + hashCode() + " closed";
        }
    }
}
	    
