/*
 *  Copyright (c) xlightweb.org, 2008 - 2009. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xlightweb.AbstractHttpConnection.IMultimodeExecutor;




/**
 * in-memory body data source
 * 
 * @author grro@xlightweb.org
 *
 */
final class FileDataSource extends NonBlockingBodyDataSource {
    
    private static final Logger LOG = Logger.getLogger(FileDataSource.class.getName());
    
    
    private static final int MAX_BUFFER_SIZE = Integer.parseInt(System.getProperty("org.xsocket.connection.filedatasource.buffersize", "16384"));
    
    private final File file;
    private final FileChannel fc;
    private final RandomAccessFile raf;
    
    private final AtomicBoolean isSuspended = new AtomicBoolean(false);
    
    private long length = 0;
    private int read = 0;
     
    
    FileDataSource(String characterEncoding, IMultimodeExecutor executor, File file) throws IOException {
        super(characterEncoding, executor);
        
        this.file = file;
        raf = new RandomAccessFile(file, "r");
        length = raf.length();
        this.fc = raf.getChannel();
        
        onRead();
    }
    
    
    /**
     * {@inheritDoc}
     */
    @Override
    protected boolean isNetworkendpoint() {
        return false;
    }


    @Override
    protected void onRead() throws IOException {
 
        if (!isSuspended()) {
            if (read >= length) {
                if (!isComplete()) {
                    raf.close();
                    fc.close();
                    setComplete();
                }
                
            } else {
                int bufferSize = (int) (length - read);
                if (bufferSize > MAX_BUFFER_SIZE) {
                    bufferSize = MAX_BUFFER_SIZE;
                }
                
                ByteBuffer buf = ByteBuffer.allocate(bufferSize);
                fc.read(buf);
                buf.flip();
                read += buf.remaining();

                if (LOG.isLoggable(Level.FINE)) {
                    LOG.fine(buf.remaining() + " bytes read from " + file.getAbsolutePath());
                }
                
                append(buf);
            }
        }
    }


    /**
     * {@inheritDoc}
     */
    @Override
    String getId() {
        return Integer.toString(hashCode());
    }

    
    /**
     * {@inheritDoc}
     */
    @Override
    boolean suspend() throws IOException {
        return false;
    }


    /**
     * {@inheritDoc}
     */
    @Override
    boolean isSuspended() {
        return isSuspended.get();
    }

    
    /**
     * {@inheritDoc}
     */
    @Override
    boolean resume() throws IOException {
        return false;
    }
  

    /**
     * {@inheritDoc}
     */
    @Override
    void onClose() {
    }

    
    /**
     * {@inheritDoc}
     */
    @Override
    void onDestroy(String reason) {
    }
}
	    
